from django.conf import settings
from rest_framework.settings import APISettings

DEFAULTS = {
    "USER_OPTIONAL_FIELDS": {
        "first_name": "rest_framework.serializers.CharField",
        "last_name": "rest_framework.serializers.CharField",
        "password": "rest_framework.serializers.CharField",
    },
    "USER_SOCIAL_AUTH_FIELDS": {
        "first_name": "rest_framework.serializers.CharField",
        "last_name": "rest_framework.serializers.CharField",
    },
    "USER_IDENTITY_FIELDS": {
        "username": "rest_framework.serializers.CharField",
        "email": "rest_framework.serializers.CharField",
        "phone_number": "phonenumber_field.serializerfields.PhoneNumberField",
    },
    "REQUIRED_AUTH_FIELDS": {},
    "OPTIONAL_AUTH_FIELDS": {
        "otp": "rest_framework.serializers.CharField",
        "password": "rest_framework.serializers.CharField",
    },
    "TEST_USER_EMAIL": None,
    "OTP_IDENTITY_UPDATE_FIELD": True,
    "OTP_DEVICE_MODELS": {
        "email": "django_otp.plugins.otp_email.models.EmailDevice",
        "totp": "django_otp.plugins.otp_totp.models.TOTPDevice",
        "sms": "otp_twilio.models.TwilioSMSDevice",
    },
    "OTP_AUTO_CREATE_ACCOUNT": True,
    "OTP_SEND_UNAUTHORIZED_USER": True,
    "SIGNUP_ALLOWED": True,
    "INVITE_ALLOWED": True,
    "SITE_LOGIN_URL": "/login/",
}

api_settings = APISettings(getattr(settings, "DF_AUTH", None), DEFAULTS)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DFAuthConfig(AppConfig):
    name = "df_auth"
    verbose_name = _("DjangoFlow Auth")

    class DFMeta:
        api_path = "auth/"



from typing import Dict, List, Tuple, Type

from django.contrib.auth.base_user import AbstractBaseUser
from django.utils.module_loading import import_string
from django_otp.models import Device

from .settings import api_settings


def get_otp_device_models() -> Dict[str, Type[Device]]:
    return {
        type_: import_string(model_path)
        for type_, model_path in api_settings.OTP_DEVICE_MODELS.items()
    }


def get_otp_device_choices() -> List[Tuple[str, str]]:
    return [(type_, type_) for type_ in api_settings.OTP_DEVICE_MODELS]


def get_otp_devices(user: AbstractBaseUser) -> List[Device]:
    devices = []
    for DeviceModel in get_otp_device_models().values():
        devices.extend(DeviceModel.objects.filter(user=user))
    return devices



import base64
from typing import Dict

from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def auth_magic_link(context: Dict) -> str:
    """
    :return: context("base_url") + base64.urlsafe_b64encode(context["username")/context("token"))
    """
    token = base64.urlsafe_b64encode(
        bytes(f"{context['username']}/{context['token']}", "utf-8")
    ).decode("utf-8")
    url = f"{context['base_url']}{token}"
    if context.get("redirect_path"):
        url += f"?redirect_path={context['redirect_path']}"
    return url



from typing import Sequence

from django.contrib import admin, messages
from django.db.models import QuerySet
from django.http import HttpRequest
from django_otp.plugins.otp_email.models import EmailDevice
from otp_twilio.models import TwilioSMSDevice

from df_auth.models import User2FA, UserRegistration
from df_auth.settings import api_settings


@admin.register(TwilioSMSDevice)
class TwilioSMSDeviceAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "number",
        "user",
    )

    search_fields = (
        "name",
        "number",
    )
    autocomplete_fields = ("user",)

    @admin.action(description="Send challenge")
    def send_challenge(
        self, request: HttpRequest, queryset: QuerySet[TwilioSMSDevice]
    ) -> None:
        for device in queryset:
            device.generate_challenge()
            messages.success(request, f"{device.number}: {device.token}")

    actions = [send_challenge]


@admin.register(EmailDevice)
class EmailDeviceAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "email",
        "user",
    )

    search_fields = (
        "name",
        "email",
    )
    autocomplete_fields = ("user",)

    @admin.action(description="Send challenge")
    def send_challenge(
        self, request: HttpRequest, queryset: QuerySet[EmailDevice]
    ) -> None:
        for device in queryset:
            device.generate_challenge()
            messages.success(request, f"{device.email}: {device.token}")

    actions = [send_challenge]


class BaseUserAdmin(admin.ModelAdmin):
    def get_search_fields(self, request: HttpRequest) -> Sequence[str]:
        return ["user__id"] + [
            f"user__{field}" for field in api_settings.USER_IDENTITY_FIELDS
        ]


@admin.register(User2FA)
class User2FAAdmin(BaseUserAdmin):
    list_display = ("user", "is_required")
    autocomplete_fields = ("user",)

    def enable(self, request: HttpRequest, queryset: QuerySet[User2FA]) -> None:
        queryset.update(is_required=True)
        messages.success(request, f"Enabled {queryset.count()} users")

    def disable(self, request: HttpRequest, queryset: QuerySet[User2FA]) -> None:
        queryset.update(is_required=False)
        messages.success(request, f"Disabled {queryset.count()} users")

    actions = [enable, disable]


@admin.register(UserRegistration)
class UserRegistrationAdmin(BaseUserAdmin):
    list_display = ("user", "is_registering", "invited_by")
    autocomplete_fields = ("user", "invited_by")



from typing import Any

from rest_framework.permissions import IsAuthenticated

from df_auth.settings import api_settings


class IsUnauthenticated(IsAuthenticated):
    def has_permission(self, *args: Any, **kwargs: Any) -> bool:
        return not super().has_permission(*args, **kwargs)


class IsUserCreateAllowed(IsAuthenticated):
    def has_permission(self, *args: Any, **kwargs: Any) -> bool:
        if super().has_permission(*args, **kwargs):
            return api_settings.INVITE_ALLOWED
        else:
            return api_settings.SIGNUP_ALLOWED



from typing import Any, Dict, Optional

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import update_last_login
from django.db.models import Model
from django.utils.module_loading import import_string
from django_otp.models import Device
from django_otp.plugins.otp_email.models import EmailDevice
from django_otp.plugins.otp_totp.models import TOTPDevice
from otp_twilio.models import TwilioSMSDevice
from rest_framework import exceptions, serializers
from rest_framework_simplejwt.settings import (
    api_settings as simplejwt_settings,
)
from social_core.exceptions import AuthCanceled, AuthForbidden
from social_django.models import DjangoStorage
from social_django.utils import load_backend

from ..exceptions import Authentication2FARequiredError
from ..models import User2FA, UserRegistration
from ..settings import api_settings
from ..strategy import DRFStrategy
from ..utils import (
    get_otp_device_choices,
    get_otp_device_models,
    get_otp_devices,
)

User = get_user_model()

AUTHENTICATION_BACKENDS = [
    import_string(backend) for backend in settings.AUTHENTICATION_BACKENDS
]


def build_fields(fields: Dict[str, str], **kwargs: Any) -> Dict[str, serializers.Field]:
    return {name: import_string(klass)(**kwargs) for name, klass in fields.items()}


def check_user_2fa(user: Optional[AbstractBaseUser], otp: Optional[str]) -> None:
    if user and hasattr(user, "user2fa") and user.user2fa.is_required:
        devices = [d for d in get_otp_devices(user) if d.confirmed]

        if not any(d.verify_token(otp) for d in devices):
            raise Authentication2FARequiredError(
                extra_data={"devices": OTPDeviceSerializer(devices, many=True).data}
            )


class EmptySerializer(serializers.Serializer):
    pass


class TokenSerializer(serializers.Serializer):
    token = serializers.CharField(read_only=True)
    token_class = simplejwt_settings.AUTH_TOKEN_CLASSES[0]
    user = None


class TokenCreateSerializer(TokenSerializer):
    @classmethod
    def get_token(cls, user: AbstractBaseUser) -> None:
        return cls.token_class.for_user(user)

    def validate(self, attrs: Dict[str, Any]) -> Dict[str, Any]:
        if not simplejwt_settings.USER_AUTHENTICATION_RULE(self.user):
            raise exceptions.AuthenticationFailed()

        token = self.get_token(self.user)  # type: ignore

        attrs["token"] = str(token)

        if simplejwt_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)  # type: ignore

        return attrs


class TokenObtainSerializer(TokenCreateSerializer):
    def validate(self, attrs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Remove empty values to pass to authenticate or send_otp
        """
        attrs = {k: v for k, v in attrs.items() if v}
        self.user = authenticate(**attrs, **self.context)  # type: ignore
        check_user_2fa(self.user, attrs.get("otp"))

        return super().validate(attrs)

    def get_fields(self) -> Dict[str, serializers.Field]:
        fields = super().get_fields()
        fields.update(
            **build_fields(
                api_settings.REQUIRED_AUTH_FIELDS,
                write_only=True,
                required=True,
                allow_blank=False,
            ),
            **build_fields(
                {
                    **api_settings.OPTIONAL_AUTH_FIELDS,
                    **api_settings.USER_IDENTITY_FIELDS,
                },
                write_only=True,
                required=False,
                allow_blank=True,
            ),
        )

        return fields


class AuthBackendSerializer(serializers.Serializer):
    backend_method_name: Optional[str] = None
    backend_extra_kwargs: Dict[str, Any] = {}

    def get_fields(self) -> Dict[str, serializers.Field]:
        return {
            **super().get_fields(),
            **build_fields(
                {
                    **api_settings.USER_IDENTITY_FIELDS,
                    **api_settings.REQUIRED_AUTH_FIELDS,
                    **api_settings.OPTIONAL_AUTH_FIELDS,
                },
                required=False,
                allow_blank=True,
            ),
        }

    def validate(self, attrs: Dict[str, Any]) -> Dict[str, Any]:
        attrs = {k: v for k, v in attrs.items() if v}
        attrs = super().validate(attrs)
        for backend in AUTHENTICATION_BACKENDS:
            if self.backend_method_name and hasattr(backend, self.backend_method_name):
                self.user = getattr(backend(), self.backend_method_name)(
                    **attrs, **self.backend_extra_kwargs, **self.context
                )
                if self.user:
                    return attrs
        raise exceptions.AuthenticationFailed(
            "Authorization backend not found", code="not_found"
        )


class OTPObtainSerializer(AuthBackendSerializer):
    backend_method_name = "generate_challenge"

    def get_fields(self) -> Dict[str, serializers.Field]:
        return {
            **super().get_fields(),
            "redirect_path": serializers.CharField(required=False, write_only=True),
        }

    def validate(self, attrs: Dict[str, Any]) -> Dict[str, Any]:
        """
        - check user auth:
          - if request.user authorized else
          - if user = authenticate(**attrs, **self.context)
          - else not authorized
        - if not authorized + OTP_SEND_UNAUTHORIZED_USER=False -> raise an error
        - send otp for the device
        """
        attrs = {k: v for k, v in attrs.items() if v}

        # check user auth
        if self.context["request"].user.is_authenticated:
            user = self.context["request"].user
        else:
            user = authenticate(**attrs, **self.context)

        # if not authorized + OTP_SEND_UNAUTHORIZED_USER=False -> raise an error
        if not user and not api_settings.OTP_SEND_UNAUTHORIZED_USER:
            raise exceptions.AuthenticationFailed(
                "Please log in to request your OTP code.",
                code="unauthorized_otp_request",
            )
        self.context["user"] = user

        # retrieve device + generate challenge
        return super().validate(attrs)


class SocialTokenObtainSerializer(TokenCreateSerializer):
    access_token = serializers.CharField(write_only=True)
    provider = serializers.ChoiceField(
        choices=[
            (backend.name, backend.name)
            for backend in AUTHENTICATION_BACKENDS
            if hasattr(backend, "name")
        ],
    )

    response = serializers.JSONField(read_only=True)

    def get_fields(self) -> Dict[str, serializers.Field]:
        return {
            **super().get_fields(),
            **build_fields(
                {
                    **api_settings.USER_SOCIAL_AUTH_FIELDS,
                    **api_settings.OPTIONAL_AUTH_FIELDS,
                },
                write_only=True,
                required=False,
                allow_blank=True,
            ),
        }

    def validate(self, attrs: Dict[str, Any]) -> Dict[str, Any]:
        request = self.context["request"]
        user = request.user if request.user.is_authenticated else None
        request.social_strategy = DRFStrategy(DjangoStorage, request)
        request.backend = load_backend(
            request.social_strategy, attrs["provider"], redirect_uri=None
        )

        try:
            self.user = request.backend.do_auth(attrs["access_token"], user=user)
        except (AuthCanceled, AuthForbidden):
            raise exceptions.AuthenticationFailed()

        update_fields = []
        for attr in api_settings.USER_SOCIAL_AUTH_FIELDS:
            if not getattr(self.user, attr, None):
                value = attrs.get(attr, None)
                if value:
                    setattr(self.user, attr, value)
                    update_fields.append(attr)
        if update_fields:
            self.user.save(update_fields=update_fields)

        check_user_2fa(self.user, attrs.get("otp"))
        return super().validate(attrs)


class OTPDeviceTypeField(serializers.ChoiceField):
    def __init__(self, **kwargs: Any) -> None:
        kwargs["source"] = "*"
        super().__init__(**kwargs)

    def to_representation(self, value: Model) -> Optional[str]:
        for type_, model in get_otp_device_models().items():
            if isinstance(value, model):
                return type_
        return None

    def to_internal_value(self, data: Any) -> Dict[str, Any]:
        return {self.field_name: data}


class OTPDeviceSerializer(serializers.Serializer):
    id = serializers.IntegerField(read_only=True)
    name = serializers.CharField(required=False)
    type = OTPDeviceTypeField(choices=get_otp_device_choices(), source="*")
    confirmed = serializers.BooleanField(read_only=True)
    extra_data = serializers.SerializerMethodField()

    def get_extra_data(self, obj: Device) -> Dict[str, str]:
        # We need `url` field for TOTP devices on `create` action
        if (
            isinstance(obj, TOTPDevice)
            and "view" in self.context
            and self.context["view"].action == "create"
        ):
            return {
                "url": obj.config_url,
            }

        return {}

    def create(self, validated_data: Dict[str, Any]) -> Device:
        device_type = validated_data.pop("type")
        DeviceModel = get_otp_device_models()[device_type]

        data = {
            "user": self.context["request"].user,
        }

        # TODO: create common interface to
        # - check if user already this device
        # - add additional fields
        # - validate if we can create device with given name (email/phone)
        if device_type == "sms":
            data["number"] = validated_data["name"]
        if device_type == "email":
            data["email"] = validated_data["name"]

        device, _ = DeviceModel.objects.get_or_create(
            **data,
            defaults={
                "confirmed": False,
                **validated_data,
            },
        )
        return device


class OTPDeviceConfirmSerializer(serializers.Serializer):
    otp = serializers.CharField(required=True, write_only=True)

    def validate_otp(self, value: str) -> str:
        if not self.instance.verify_token(value):
            raise serializers.ValidationError("Invalid OTP code")
        return value

    def update(self, instance: Device, validated_data: Dict[str, Any]) -> Device:
        instance.confirmed = True
        instance.save()

        if api_settings.OTP_IDENTITY_UPDATE_FIELD:
            # TODO: create a common interface for this
            if isinstance(instance, EmailDevice):
                instance.user.email = instance.name
            elif isinstance(instance, TwilioSMSDevice):
                instance.user.phone_number = instance.number
            instance.user.save()

        return instance


class UserIdentitySerializer(serializers.Serializer):
    def get_fields(self) -> Dict[str, serializers.Field]:
        fields = build_fields(
            {
                **api_settings.USER_OPTIONAL_FIELDS,
                **api_settings.USER_IDENTITY_FIELDS,
            },
            required=False,
            allow_blank=True,
        )

        if "password" in fields:
            fields["password"].write_only = True

        return fields

    def validate_email(self, value: Optional[str]) -> Optional[str]:
        if not value:
            return None
        # TODO: check for black list

        # If User has no confirmed EmailDevice on update
        if (
            self.instance
            and not self.instance.emaildevice_set.filter(
                email=value, confirmed=True
            ).exists()
        ):
            raise serializers.ValidationError("You need to confirm your email first.")

        if (
            self.instance is None
            and "email" in api_settings.USER_IDENTITY_FIELDS
            and value is not None
            and User.objects.filter(email=value).exists()
        ):
            raise serializers.ValidationError("User with this email already exists.")

        return User.objects.normalize_email(value)

    def validate_username(self, value: Optional[str]) -> Optional[str]:
        if not value:
            return None
        # TODO: check for black list
        return value

    def validate_phone_number(self, value: Optional[str]) -> Optional[str]:
        if not value:
            return None

        # TODO: check for black list
        if (
            self.instance
            and not self.instance.twiliosmsdevice_set.filter(
                number=value, confirmed=True
            ).exists()
        ):
            raise serializers.ValidationError(
                "You need to confirm your phone number first."
            )

        if (
            self.instance is None
            and "phone_number" in api_settings.USER_IDENTITY_FIELDS
            and value is not None
            and User.objects.filter(phone_number=value).exists()
        ):
            raise serializers.ValidationError(
                "User with this phone number already exists."
            )

        return value

    def update(self, instance: User, validated_data: Dict[str, Any]) -> User:
        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        instance.save()
        return instance

    def create(self, validated_data: Any) -> User:
        if not validated_data.get("username") and getattr(User, "username", False):
            for field in api_settings.USER_IDENTITY_FIELDS:
                if validated_data.get(field):
                    validated_data["username"] = validated_data[field]
                    break

        user = User(**validated_data)
        if validated_data.get("password"):
            user.set_password(validated_data["password"])
        user.save()

        request_user = self.context["request"].user
        UserRegistration.objects.create(
            user=user,
            invited_by=request_user if request_user.is_authenticated else None,
        )

        # TODO: create common interface
        if getattr(User, "email", False) and user.email:  # type: ignore
            EmailDevice.objects.create(
                user=user, email=user.email, confirmed=False, name=user.email  # type: ignore
            )

        if getattr(User, "phone_number", False) and user.phone_number:  # type: ignore
            TwilioSMSDevice.objects.create(
                user=user,
                number=user.phone_number,
                confirmed=False,
                name=user.phone_number,
            )

        return user


class ChangePasswordSerializer(serializers.Serializer):
    old_password = serializers.CharField(required=True, write_only=True)
    new_password = serializers.CharField(required=True, write_only=True)

    def validate_old_password(self, value: str) -> str:
        if not self.instance.check_password(value):
            raise serializers.ValidationError("Old password is incorrect.")
        return value

    def update(self, instance: User, validated_data: Dict[str, Any]) -> User:
        instance.set_password(validated_data["new_password"])
        instance.save()
        return instance


class User2FASerializer(serializers.ModelSerializer):
    class Meta:
        model = User2FA
        fields = ["is_required"]



from django.conf import settings
from django.db import models


class UserOneToOneMixin(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True


class User2FA(UserOneToOneMixin):
    is_required = models.BooleanField(default=False)

    class Meta:
        verbose_name = "User 2FA"
        verbose_name_plural = "User 2FA"


class UserRegistration(UserOneToOneMixin):
    is_registering = models.BooleanField(default=True)
    invited_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="invitees",
        null=True,
        blank=True,
    )



# Generated by Django 4.2.4 on 2023-10-05 12:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("df_auth", "0002_alter_user2fa_user_userregistration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userregistration",
            name="is_registering",
            field=models.BooleanField(default=True),
        ),
    ]



from typing import Any

from rest_framework.request import Request
from social_django.storage import BaseDjangoStorage
from social_django.strategy import DjangoStrategy


class DRFStrategy(DjangoStrategy):
    def __init__(self, storage: BaseDjangoStorage, request: Request) -> None:
        self.request = request
        super().__init__(storage, request)

    def request_data(self, merge: bool = True) -> Any:
        return self.request.data



from typing import Any, Optional, Type

from df_api_drf.resolvers import client_url
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.http import HttpRequest
from django_otp.models import SideChannelDevice
from django_otp.plugins.otp_email.models import EmailDevice
from otp_twilio.models import TwilioSMSDevice

from .exceptions import (
    UserDoesNotExistError,
    UserInactiveError,
    WrongOTPError,
)
from .settings import api_settings

User = get_user_model()


class TestEmailBackend(ModelBackend):
    def authenticate(self, request: Optional[HttpRequest], **kwargs: Any) -> Optional[User]:  # type: ignore
        if (
            api_settings.TEST_USER_EMAIL
            and kwargs.get("email") == api_settings.TEST_USER_EMAIL
        ):
            return User._default_manager.get(email=api_settings.TEST_USER_EMAIL)

        return None


class BaseOTPBackend(ModelBackend):
    identity_field: str
    device_identity_field: str
    DeviceModel: Type[SideChannelDevice]

    def update_user_identity_field(self, device: SideChannelDevice) -> None:
        if api_settings.OTP_IDENTITY_UPDATE_FIELD:
            user = device.user
            setattr(
                user, self.identity_field, getattr(device, self.device_identity_field)
            )
            user.save()

    def generate_challenge(
        self, request: HttpRequest, user: Optional[User], **kwargs: Any
    ) -> Optional[User]:
        """
        - if user not authorized:
            - find User by some identity field: email/phone.
                - if not found:
                    - if OTP_AUTO_CREATE_ACCOUNT -> create a User with form data
                    - else: raise an error "No user, please register"

        - Create active device if does not exist
        - send otp for the device
        """
        if not kwargs.get(self.identity_field):
            return None

        if not user:
            user = User.objects.filter(
                **{self.identity_field: kwargs.get(self.identity_field)}
            ).first()
            if not user:
                if api_settings.OTP_AUTO_CREATE_ACCOUNT and api_settings.SIGNUP_ALLOWED:
                    user = User.objects.create(
                        **{
                            k: v
                            for k, v in kwargs.items()
                            if k
                            in [
                                *api_settings.USER_OPTIONAL_FIELDS,
                                *api_settings.USER_IDENTITY_FIELDS,
                            ]
                        }
                    )
                else:
                    raise UserDoesNotExistError()
        if not user.is_active:
            raise UserInactiveError()

        device, _ = self.DeviceModel.objects.get_or_create(
            user=user,
            **{self.device_identity_field: kwargs.get(self.identity_field)},
            defaults={
                "name": kwargs.get("name", kwargs.get(self.identity_field)),
                "confirmed": True,  # Because User already has this device as identity field
            },
        )

        self.send_challenge(device, request, **kwargs)
        return device.user

    def send_challenge(
        self, device: SideChannelDevice, request: HttpRequest, **kwargs: Any
    ) -> None:
        device.generate_challenge()

    def authenticate(self, request: Optional[HttpRequest], **kwargs: Any) -> Optional[User]:  # type: ignore
        """
        Check OTP and authenticate User
        """
        if not kwargs.get(self.identity_field) or not kwargs.get("otp"):
            return None

        user = User.objects.filter(
            **{self.identity_field: kwargs.get(self.identity_field)}, is_active=True
        ).first()

        if not user:
            return None

        device = self.DeviceModel.objects.filter(
            user=user, **{self.device_identity_field: kwargs.get(self.identity_field)}
        ).first()
        if device is None:
            return None

        if not device.verify_token(kwargs.get("otp")):
            raise WrongOTPError()

        return device.user


class EmailOTPBackend(BaseOTPBackend):
    identity_field = "email"
    device_identity_field = "email"
    DeviceModel = EmailDevice

    def send_challenge(
        self, device: EmailDevice, request: HttpRequest, **kwargs: Any
    ) -> None:
        device.generate_challenge(
            {
                "username": device.email,
                "base_url": client_url(request=request, user=device.user)
                + api_settings.SITE_LOGIN_URL,
                "redirect_path": kwargs.get("redirect_path", ""),
            }
        )


class TwilioSMSOTPBackend(BaseOTPBackend):
    identity_field = "phone_number"
    device_identity_field = "number"
    DeviceModel = TwilioSMSDevice



from typing import Any, Optional

from df_remote_config.handlers import DefaultHandler
from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.utils.module_loading import import_string

from df_auth.backends import BaseOTPBackend

AUTHENTICATION_BACKENDS = [
    import_string(backend) for backend in settings.AUTHENTICATION_BACKENDS
]

auth_schema = {
    "type": "object",
    "properties": {
        "providers": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "button_text": {
                        "type": "string",
                    },
                    "redirect_uri": {
                        "type": "object",
                        "properties": {
                            "web": {"type": "string"},
                            "mobile": {"type": "string"},
                        },
                    },
                },
            },
        },
        "otp": {
            "type": "object",
            "properties": {
                "enabled": {
                    "type": "boolean",
                },
            },
        },
        "email_password": {
            "type": "object",
            "properties": {
                "enabled": {
                    "type": "boolean",
                },
            },
        },
    },
    "required": ["providers", "otp", "email_password"],
}


class AuthHandler(DefaultHandler):
    def get_part_data(self, part: Optional[Any]) -> dict:
        data = super().get_part_data(part)

        enabled_providers = [
            backend.name
            for backend in AUTHENTICATION_BACKENDS
            if hasattr(backend, "name")
        ]

        for provider in data["providers"]:
            if "enabled" not in data["providers"][provider]:
                data["providers"][provider]["enabled"] = provider in enabled_providers
            data["providers"][provider]["client_id"] = getattr(
                settings, f"SOCIAL_AUTH_{provider.upper()}_KEY", None
            )

        for provider in AUTHENTICATION_BACKENDS:
            if (
                issubclass(provider, ModelBackend)
                and "enabled" not in data["email_password"]
            ):
                data["email_password"].update(
                    {
                        "enabled": True,
                    }
                )
            if issubclass(provider, BaseOTPBackend) and "enabled" not in data["otp"]:
                data["otp"].update(
                    {
                        "enabled": True,
                    }
                )

        return data



"""Djangoflow URL Configuration

Add these to your root URLconf:
    urlpatterns = [
        ...
        path('auth/', include('df_auth.urls'))
    ]

"""
from rest_framework.routers import DefaultRouter

from .viewsets import (
    OtpDeviceViewSet,
    OTPViewSet,
    SocialTokenViewSet,
    TokenViewSet,
    UserViewSet,
)

router = DefaultRouter()
router.register("token", TokenViewSet, basename="token")
router.register("users", UserViewSet, basename="users")
router.register("otp", OTPViewSet, basename="otp")
router.register("otp-devices", OtpDeviceViewSet, basename="otp-devices")

router.register("social", SocialTokenViewSet, basename="social")

urlpatterns = router.urls



from typing import Any, Iterable, List, Type

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django_otp.models import Device
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import permissions, response, status, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import BasePermission
from rest_framework.settings import import_string
from rest_framework_simplejwt.settings import (
    api_settings as simple_jwt_settings,
)

from ..exceptions import (
    DfAuthValidationError,
)
from ..models import User2FA
from ..permissions import IsUnauthenticated, IsUserCreateAllowed
from ..utils import get_otp_device_models, get_otp_devices
from .serializers import (
    ChangePasswordSerializer,
    OTPDeviceConfirmSerializer,
    OTPDeviceSerializer,
    OTPObtainSerializer,
    SocialTokenObtainSerializer,
    TokenObtainSerializer,
    TokenSerializer,
    User2FASerializer,
    UserIdentitySerializer,
)


class ValidationOnlyCreateViewSet(viewsets.GenericViewSet):
    def create(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        return response.Response(serializer.data, status=status.HTTP_200_OK)


class TokenViewSet(ValidationOnlyCreateViewSet):
    serializer_class = TokenObtainSerializer
    response_serializer_class = TokenSerializer
    permission_classes = (permissions.AllowAny,)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=import_string(simple_jwt_settings.TOKEN_REFRESH_SERIALIZER),
    )
    def refresh(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return self.create(request, *args, **kwargs)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=import_string(simple_jwt_settings.TOKEN_VERIFY_SERIALIZER),
    )
    def verify(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return self.create(request, *args, **kwargs)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=import_string(simple_jwt_settings.TOKEN_BLACKLIST_SERIALIZER),
    )
    def blacklist(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        if "rest_framework_simplejwt.token_blacklist" not in settings.INSTALLED_APPS:
            raise NotImplementedError

        return self.create(request, *args, **kwargs)


class OTPViewSet(ValidationOnlyCreateViewSet):
    throttle_scope = "otp"
    serializer_class = OTPObtainSerializer
    permission_classes = (permissions.AllowAny,)


class SocialTokenViewSet(ValidationOnlyCreateViewSet):
    serializer_class = SocialTokenObtainSerializer
    response_serializer_class = TokenSerializer
    permission_classes = (IsUnauthenticated,)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=SocialTokenObtainSerializer,
        permission_classes=(permissions.IsAuthenticated,),
    )
    def connect(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return self.create(request, *args, **kwargs)


otp_device_detail_params = [
    OpenApiParameter(
        name="type",
        type=OpenApiTypes.STR,
        location=OpenApiParameter.QUERY,
        description="OTP Device type",
        required=True,
    )
]


class OtpDeviceViewSet(
    viewsets.GenericViewSet,
    viewsets.mixins.ListModelMixin,
    viewsets.mixins.DestroyModelMixin,
    viewsets.mixins.RetrieveModelMixin,
    viewsets.mixins.CreateModelMixin,
):
    throttle_scope = "otp"
    serializer_class = OTPDeviceSerializer
    permission_classes = (permissions.IsAuthenticated,)

    def get_queryset(self) -> List[Device]:
        return get_otp_devices(self.request.user)

    def get_device_model(self) -> Type[Device]:
        device_type = self.request.GET.get("type")
        otp_device_models = get_otp_device_models()
        if device_type not in otp_device_models:
            raise DfAuthValidationError(
                f"Invalid device type. Must be one of {', '.join(otp_device_models.keys())}"
            )
        return otp_device_models[device_type]

    def get_object(self) -> Device:
        return self.get_device_model().objects.get(
            user=self.request.user, pk=self.kwargs["pk"]
        )

    @extend_schema(parameters=otp_device_detail_params)
    @action(
        methods=["post"],
        detail=True,
        serializer_class=OTPDeviceConfirmSerializer,
    )
    def confirm(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        serializer = OTPDeviceConfirmSerializer(
            data=request.data, instance=self.get_object()
        )
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response({})

    @extend_schema(parameters=otp_device_detail_params)
    def destroy(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return super().destroy(request, *args, **kwargs)

    @extend_schema(parameters=otp_device_detail_params)
    def retrieve(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return super().retrieve(request, *args, **kwargs)


class UserViewSet(
    viewsets.GenericViewSet,
    viewsets.mixins.RetrieveModelMixin,
    viewsets.mixins.CreateModelMixin,
    viewsets.mixins.UpdateModelMixin,
):
    serializer_class = UserIdentitySerializer
    permission_classes = (permissions.IsAuthenticated,)
    http_method_names = ["get", "post", "patch"]

    def get_permissions(self) -> Iterable[BasePermission]:
        if self.action == "create":
            return (IsUserCreateAllowed(),)
        return super().get_permissions()

    def get_object(self) -> Any:
        return self.request.user

    @action(
        detail=True,
        methods=["POST"],
        serializer_class=ChangePasswordSerializer,
        url_path="set-password",
    )
    def set_password(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        return super().update(request, *args, **kwargs)

    @action(
        detail=True,
        methods=["GET", "PATCH"],
        serializer_class=User2FASerializer,
        url_path="two-fa",
    )
    def two_fa(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        instance = User2FA.objects.get_or_create(user=self.get_object())[0]
        if request.method == "GET":
            serializer = self.get_serializer(instance)
        else:
            serializer = self.get_serializer(instance, data=request.data, partial=True)
            serializer.is_valid(raise_exception=True)
            serializer.save()
        return response.Response(serializer.data)



from datetime import timedelta

DF_AUTH_APPS_BASE = [
    "df_auth",
    "rest_framework_simplejwt",
]

DF_AUTH_APPS_OTP = [
    "django_otp",
    "django_otp.plugins.otp_email",
    "django_otp.plugins.otp_totp",
    "django_otp.plugins.otp_static",
    "otp_twilio",
]

DF_AUTH_APPS_SOCIAL = [
    "social_django",
]

# Most common use case
DF_AUTH_INSTALLED_APPS = [
    *DF_AUTH_APPS_BASE,
    *DF_AUTH_APPS_OTP,
    *DF_AUTH_APPS_SOCIAL,
]

SIMPLE_JWT = {
    "ACCESS_TOKEN_LIFETIME": timedelta(days=14),
    "AUTH_HEADER_TYPES": ("Bearer",),
    "AUTH_TOKEN_CLASSES": ("rest_framework_simplejwt.tokens.SlidingToken",),
    "BLACKLIST_AFTER_ROTATION": False,
    "JTI_CLAIM": "jti",
    "REFRESH_TOKEN_LIFETIME": timedelta(days=14),
    "ROTATE_REFRESH_TOKENS": True,
    "SLIDING_TOKEN_LIFETIME": timedelta(days=14),
    "SLIDING_TOKEN_REFRESH_EXP_CLAIM": "refresh_exp",
    "SLIDING_TOKEN_REFRESH_LIFETIME": timedelta(days=14),
    "TOKEN_TYPE_CLAIM": "sliding",
    "USER_ID_CLAIM": "user_id",
    "USER_ID_FIELD": "id",
    "TOKEN_OBTAIN_SERIALIZER": "rest_framework_simplejwt.serializers.TokenObtainSlidingSerializer",
    "TOKEN_REFRESH_SERIALIZER": "rest_framework_simplejwt.serializers.TokenRefreshSlidingSerializer",
}

SOCIAL_AUTH_PIPELINE = [
    # Get the information we can about the user and return it in a simple
    # format to create the user instance later. On some cases the details are
    # already part of the auth response from the provider, but sometimes this
    # could hit a provider API.
    "social_core.pipeline.social_auth.social_details",
    # Get the social uid from whichever service we're authing thru. The uid is
    # the unique identifier of the given user in the provider.
    "social_core.pipeline.social_auth.social_uid",
    # Verifies that the current auth process is valid within the current
    # project, this is where emails and domains whitelists are applied (if
    # defined).
    # 'social_core.pipeline.social_auth.auth_allowed',
    # Checks if the current social-account is already associated in the site.
    "social_core.pipeline.social_auth.social_user",
    # Make up a username for this person, appends a random string at the end if
    # there's any collision.
    "social_core.pipeline.user.get_username",
    # Send a validation email to the user to verify its email address.
    # 'social_core.pipeline.mail.mail_validation',
    # Associates the current social details with another user account with
    # a similar email address.
    "social_core.pipeline.social_auth.associate_by_email",
    # Create a user account if we haven't found one yet.
    "social_core.pipeline.user.create_user",
    # Create the record that associated the social account with this user.
    "social_core.pipeline.social_auth.associate_user",
    # Populate the extra_data field in the social record with the values
    # specified by settings (and the default ones like access_token, etc).
    "social_core.pipeline.social_auth.load_extra_data",
    # Update the user record with any changed info from the auth service.
    "social_core.pipeline.user.user_details",
]



from df_api_drf.exceptions import ExtraDataAPIException
from django.utils.translation import gettext_lazy as _
from rest_framework import status
from rest_framework.exceptions import (
    PermissionDenied,
    ValidationError,
)


class DfAuthValidationError(ValidationError):
    """
    This is a base exception for custom validation errors
    """

    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _("Authentication error")


class WrongOTPError(DfAuthValidationError):
    """
    This exception is used when token for otp is not verified
    """

    default_detail = _("Wrong or expired one-time password")
    default_code = "wrong_otp"


class UserAlreadyExistError(DfAuthValidationError):
    """
    This exception is used when user already exists
    """

    default_detail = _("User with this identity already exists, try logging in")
    default_code = "user_already_exists"


class UserDoesNotExistError(DfAuthValidationError):
    """
    This exception is used when user already exists
    """

    default_detail = _("User with this identity does not exist, try signup instead")
    default_code = "user_does_not_exist"


class UserInactiveError(DfAuthValidationError):
    """
    This exception is used when user already exists
    """

    default_detail = _("Your account was deactivated. Please contact support")
    default_code = "user_inactive"


class DeviceTakenError(DfAuthValidationError):
    """
    This exception is used when device is already registered
    """

    default_detail = _("This device is already taken, unlink it first")
    default_code = "device_already_taken"


class InvalidPhoneNumberError(DfAuthValidationError):
    """
    This exception is used when phone number is not valid
    """

    default_detail = _("Invalid phone number")
    default_code = "invalid_phone_number"


class DeviceDoesNotExistError(DfAuthValidationError):
    """
    This exception is used when device is not registered
    """

    default_detail = _("Device does not exist")
    default_code = "device_does_not_exists"


class LastDeviceError(DfAuthValidationError):
    """
    This exception is used when there is no device registered with user
    """

    default_detail = _("Cannot remove the last device")
    default_code = "last_device_error"


class Authentication2FARequiredError(ExtraDataAPIException):
    default_detail = "2FA is required for this user."
    default_code = "2fa_required"
    status_code = status.HTTP_401_UNAUTHORIZED


class SignupNotAllowedError(PermissionDenied):
    default_detail = "Signup is disabled."
    default_code = "signup_not_allowed"



from typing import Any

from django.contrib.auth import get_user_model
from django.contrib.auth.base_user import BaseUserManager


class UserManager(BaseUserManager):
    def create_superuser(self, **extra_fields: Any) -> Any:
        """Create and save a SuperUser with the given email and password."""
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return self._create_user(**extra_fields)

    def _create_user(self, **extra_fields: Any) -> Any:
        """Create and save a User with the given email and password."""
        User = get_user_model()
        password = extra_fields.pop("password", None)

        if not extra_fields.get(User.USERNAME_FIELD):
            raise ValueError(f"The given {User.USERNAME_FIELD} must be set")
        if User.EMAIL_FIELD in extra_fields:
            extra_fields[User.EMAIL_FIELD] = self.normalize_email(
                extra_fields[User.EMAIL_FIELD]
            )

        user = self.model(**extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, **extra_fields: Any) -> Any:
        """Create and save a regular User with the given email and password."""
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(**extra_fields)
